// This file is part of Tibbo Device Monitor (R)
// Tibbo Technology Inc (C) 2004-2015. All rights reserved
// Author: Vladimir Gladkov

#pragma once

#ifndef __cplusplus
typedef enum dm_NotifyCode                    dm_NotifyCode;
typedef struct dm_Stats                       dm_Stats;
typedef struct dm_DeviceInfo                  dm_DeviceInfo;
typedef struct dm_DeviceInfoList              dm_DeviceInfoList;
typedef struct dm_ConnectParams               dm_ConnectParams;
typedef struct dm_NotifyHdr                   dm_NotifyHdr;
typedef struct dm_CreateNotifyParams          dm_CreateNotifyParams;
typedef struct dm_CreateNamedPipeNotifyParams dm_CreateNamedPipeNotifyParams;
typedef struct dm_CloseNotifyParams           dm_CloseNotifyParams;
typedef struct dm_ReadWriteNotifyParams       dm_ReadWriteNotifyParams;
typedef struct dm_IoctlNotifyParams           dm_IoctlNotifyParams;
typedef union dm_NotifyParams                 dm_NotifyParams;
typedef union dm_NotifyParamsPtr              dm_NotifyParamsPtr;
#endif

//.............................................................................

enum
{
	dm_ConnectionCountLimit      = 16,              // no more than 16 connections to a device
	dm_DefPendingNotifySizeLimit = 1 * 1024 * 1024, // drop notifications if application is not fast enough to pick'em up
	dm_NotifyHdrSignature        = 'nomt',
	dm_DeviceType                = FILE_DEVICE_UNKNOWN,
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

#define DM_MASTER_DEVICE_NAMEA        "tdevmon-master"
#define DM_BEACON_DEVICE_NAMEA        "tdevmon-beacon"
#define DM_CORE_DRIVER_FILE_NAMEA     "tdevmonc.sys"
#define DM_CORE_SERVICE_NAMEA         "tdevmonc"
#define DM_CORE_SERVICE_DISPLAY_NAMEA "Tibbo Device Monitor kernel-mode core service"
#define DM_CORE_SERVICE_DESCRIPTIONA  "Provides core monitoring capabilities for Tibbo IO Ninja"
#define DM_PNP_DRIVER_FILE_NAMEA      "tdevmonp.sys"
#define DM_PNP_SERVICE_NAMEA          "tdevmonp"
#define DM_PNP_SERVICE_DISPLAY_NAMEA  "Tibbo Device Monitor kernel-mode PNP filter service"
#define DM_PNP_SERVICE_DESCRIPTIONA   "Provides early-attach to PNP devices for Tibbo IO Ninja"

#define DM_MASTER_DEVICE_NAMEW        _CRT_WIDE (DM_MASTER_DEVICE_NAMEA)
#define DM_BEACON_DEVICE_NAMEW        _CRT_WIDE (DM_BEACON_DEVICE_NAMEA)
#define DM_CORE_DRIVER_FILE_NAMEW     _CRT_WIDE (DM_CORE_DRIVER_FILE_NAMEA)
#define DM_CORE_SERVICE_NAMEW         _CRT_WIDE (DM_CORE_SERVICE_NAMEA)
#define DM_CORE_SERVICE_DISPLAY_NAMEW _CRT_WIDE (DM_CORE_SERVICE_DISPLAY_NAMEA)
#define DM_CORE_SERVICE_DESCRIPTIONW  _CRT_WIDE (DM_CORE_SERVICE_DESCRIPTIONA)
#define DM_PNP_DRIVER_FILE_NAMEW      _CRT_WIDE (DM_PNP_DRIVER_FILE_NAMEA)
#define DM_PNP_SERVICE_NAMEW          _CRT_WIDE (DM_PNP_SERVICE_NAMEA)
#define DM_PNP_SERVICE_DISPLAY_NAMEW  _CRT_WIDE (DM_PNP_SERVICE_DISPLAY_NAMEA)
#define DM_PNP_SERVICE_DESCRIPTIONW   _CRT_WIDE (DM_PNP_SERVICE_DESCRIPTIONA)

#define DM_MASTER_DEVICE_NAME         _T (DM_MASTER_DEVICE_NAMEA)
#define DM_BEACON_DEVICE_NAME         _T (DM_BEACON_DEVICE_NAMEA)
#define DM_CORE_DRIVER_FILE_NAME      _T (DM_CORE_DRIVER_FILE_NAMEA)
#define DM_CORE_SERVICE_NAME          _T (DM_CORE_SERVICE_NAMEA)
#define DM_CORE_SERVICE_DISPLAY_NAME  _T (DM_CORE_SERVICE_DISPLAY_NAMEA)
#define DM_CORE_SERVICE_DESCRIPTION   _T (DM_CORE_SERVICE_DESCRIPTIONA)
#define DM_PNP_DRIVER_FILE_NAME       _T (DM_PNP_DRIVER_FILE_NAMEA)
#define DM_PNP_SERVICE_NAME           _T (DM_PNP_SERVICE_NAMEA)
#define DM_PNP_SERVICE_DISPLAY_NAME   _T (DM_PNP_SERVICE_DISPLAY_NAMEA)
#define DM_PNP_SERVICE_DESCRIPTION    _T (DM_PNP_SERVICE_DESCRIPTIONA)

//.............................................................................

#define DM_CTL_CODE_EX(code, method) \
	CTL_CODE (dm_DeviceType, 0x800 + code, method, FILE_ANY_ACCESS)

#define DM_CTL_CODE(code) \
	DM_CTL_CODE_EX(code, METHOD_BUFFERED) 

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

#define IOCTL_DM_GET_VERSION            DM_CTL_CODE (1)
#define IOCTL_DM_GET_DESCRIPTION        DM_CTL_CODE (2)
#define IOCTL_DM_GET_BUILD_TIME         DM_CTL_CODE (3)
#define IOCTL_DM_GET_DEVICE_INFO_LIST   DM_CTL_CODE (4)
#define IOCTL_DM_GET_TARGET_DEVICE_INFO DM_CTL_CODE (5)
#define IOCTL_DM_IS_CONNECTED           DM_CTL_CODE (6)
#define IOCTL_DM_CONNECT                DM_CTL_CODE (7)
#define IOCTL_DM_DISCONNECT             DM_CTL_CODE (8)
#define IOCTL_DM_STOP                   DM_CTL_CODE (9)

#if (DBG || defined _DEBUG || defined _AXL_DEBUG)
#	define IOCTL_DM_TEST                DM_CTL_CODE (50)
#endif

//.............................................................................

enum dm_DeviceFlag
{
	dm_DeviceFlag_RemotePnpFilter = 0x01,
	dm_DeviceFlag_PnpDevice       = 0x02,
	dm_DeviceFlag_FilterConnected = 0x04,
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct dm_DeviceInfo
{
	ULONG m_flags; // dm_DeviceFlag combination
	ULONG m_type;  // FILE_DEVICE_xxx
	ULONG m_connectionCount;
	ULONG m_requestCount;
	ULONG m_nameLength;
	
	// followed by a UTF16 device name;
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct dm_DeviceInfoList
{
	ULONG m_deviceCount;
	ULONG m_deviceInfoSize;

	// followed by dm_DeviceInfo x m_deviceCount
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct dm_ConnectParams
{
	ULONG m_pendingNotifySizeLimit; // 0 -- use dm_DefPendingNotifySizeLimit
	
	// followed by a UTF16 device name (full device name or a symbolic link)
};

//.............................................................................

enum dm_NotifyCode
{
	dm_NotifyCode_Undefined = 0,
	dm_NotifyCode_PnpStartDevice,
	dm_NotifyCode_PnpStopDevice,
	dm_NotifyCode_PnpRemoveDevice,
	dm_NotifyCode_FastIoDetachDevice,
	dm_NotifyCode_Create,
	dm_NotifyCode_CreateNamedPipe,
	dm_NotifyCode_Close,
	dm_NotifyCode_Read,
	dm_NotifyCode_Write,
	dm_NotifyCode_Ioctl,
	dm_NotifyCode_InternalIoctl,
	dm_NotifyCode__Count,
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

enum dm_NotifyFlag
{
	dm_NotifyFlag_InsufficientBuffer = 0x01, // buffer is not big enough, resize and try again
	dm_NotifyFlag_DataDropped        = 0x02, // one or more notifications after this one were dropped
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct dm_NotifyHdr
{
	UINT m_signature;
	USHORT m_code;
	USHORT m_flags;
	LONG m_ntStatus;
	ULONG m_paramsSize;

	// followed by params
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct dm_CreateNotifyParams
{
	ULONGLONG m_fileId;
	UINT m_options;
	UINT m_desiredAccess;
	UINT m_shareAccess;
	UINT m_fileAttributes;
	ULONG m_fileNameLength;
	ULONG _m_padding;

	// followed by UTF16 file name
};

#ifdef __cplusplus
struct dm_CreateNamedPipeNotifyParams: dm_CreateNotifyParams {
#else
struct dm_CreateNamedPipeNotifyParams { dm_CreateNotifyParams;
#endif

	UINT m_pipeType;
	UINT m_readMode;
	UINT m_completionMode;
	ULONG m_maxInstanceCount;
	ULONG m_inBoundQuota;
	ULONG m_outBoundQuota;
	ULONGLONG m_defaultTimeout;

	// followed by UTF16 pipe name
};

struct dm_CloseNotifyParams
{
	ULONGLONG m_fileId;
};

struct dm_ReadWriteNotifyParams
{
	ULONGLONG m_fileId;
	ULONGLONG m_offset;
	UINT m_bufferSize;
	UINT m_dataSize;

	// followed by read/write data
};

struct dm_IoctlNotifyParams
{
	ULONGLONG m_fileId;
	UINT m_ioctlCode;
	ULONG m_inDataSize;
	ULONG m_outBufferSize;
	ULONG m_outDataSize;

	// followed by in-data
	// followed by out-data
};

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

union dm_NotifyParams
{
	dm_CreateNotifyParams m_createParams;
	dm_CreateNamedPipeNotifyParams m_createNamedPipeParams;
	dm_CloseNotifyParams m_closeParams;
	dm_ReadWriteNotifyParams m_readWriteParams;
	dm_IoctlNotifyParams m_ioctlParams;
};

union dm_NotifyParamsPtr
{
	VOID* m_params;
	dm_CreateNotifyParams* m_createParams;
	dm_CreateNamedPipeNotifyParams* m_createNamedPipeParams;
	dm_CloseNotifyParams* m_closeParams;
	dm_ReadWriteNotifyParams* m_readWriteParams;
	dm_IoctlNotifyParams* m_ioctlParams;
};

#ifdef _NTDDK_

//.............................................................................

#ifndef __cplusplus
typedef struct dm_CoreDriverDispatch dm_CoreDriverDispatch;
typedef struct dm_PnpDriverDispatch  dm_PnpDriverDispatch;
#endif

//.............................................................................

// these are sent as internal IOCTLs

#define IOCTL_DM_CORE_TO_PNP_DRIVER_CONNECT DM_CTL_CODE_EX (100, METHOD_NEITHER)
#define IOCTL_DM_PNP_TO_CORE_DRIVER_CONNECT DM_CTL_CODE_EX (101, METHOD_NEITHER)

//.............................................................................

typedef 
VOID
dm_DisconnectPnpDriverFunc ();

typedef 
NTSTATUS
dm_CreateFilterDeviceFunc (
	DEVICE_OBJECT** filterDeviceObject,
	DEVICE_OBJECT* targetDeviceObject,
	const WCHAR* targetDeviceName
	);

typedef 
VOID
dm_AddRefFilterDeviceFunc (DEVICE_OBJECT* deviceObject);

typedef 
VOID
dm_ReleaseFilterDeviceFunc (DEVICE_OBJECT* deviceObject);

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct dm_CoreDriverDispatch
{
	dm_DisconnectPnpDriverFunc* m_disconnectPnpDriver;
	dm_CreateFilterDeviceFunc* m_createFilterDevice;
	dm_AddRefFilterDeviceFunc* m_addRefFilterDevice;
	dm_ReleaseFilterDeviceFunc* m_releaseFilterDevice;
};

//.............................................................................

typedef 
VOID
dm_DisconnectCoreDriverFunc ();

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct dm_PnpDriverDispatch
{
	dm_DisconnectCoreDriverFunc* m_disconnectCoreDriver;
};

//.............................................................................

#endif
